#include <windows.h>
#include <d3d10.h>
#include <d3dx10.h>
#include "resource.h"

/* create a structure to store the triangles vertex coordinates */
struct TriangleVertex
{
    D3DXVECTOR3 Coordinate;  
};

/* initialize the instance handle, g_hInstance,
   the handle to the window, g_hWindow, etc. */
HWND                    g_hWindow = NULL;
HINSTANCE               g_hInstance = NULL;  
D3D10_DRIVER_TYPE       g_dx10DriverType = D3D10_DRIVER_TYPE_NULL;
ID3D10Device*           g_pd3d10Device = NULL;
IDXGISwapChain*         g_pdx10SwapChain = NULL;
ID3D10EffectTechnique*  g_pd3d10EffectTechnique = NULL;
ID3D10Buffer*           g_pd3d10VertexBuffer = NULL;
ID3D10InputLayout*      g_pd3d10InputVertexLayout = NULL;
ID3D10Effect*           g_pd3d10Effect = NULL;
ID3D10RenderTargetView* g_pd3d10RenderTargetView = NULL;

/* function prototypes */
HRESULT InitDevice();
void Clean();
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
HRESULT InitWindow(HINSTANCE hInstance, int nCmdShow );
void DrawTriangle();

/* initialize the message processing loop elements and the idle time required
   for the rendering of our triangle via winMain, the application entry point */
int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow)
{
	if((FAILED(InitWindow(hInstance, nCmdShow)))||(FAILED(InitDevice())))
	{
		Clean(); //do device cleanup  freeing resources
		return 0;
	}

	MSG msg = {0};
	while(WM_QUIT != msg.message)
	{
		if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
		{
			DrawTriangle(); //render our triangle
		}
	}
	Clean();
	return (int)msg.wParam;
}

/* register the class and create the output window */
HRESULT InitWindow(HINSTANCE hInstance, int nCmdShow)
{
	/* register the window class */
	///////////////////////////////
	WNDCLASSEX wndcex;
	wndcex.cbSize 			= sizeof(WNDCLASSEX);
	wndcex.style 			= CS_HREDRAW | CS_VREDRAW;
	wndcex.lpfnWndProc 		= WndProc;
	wndcex.cbClsExtra 		= 0;
	wndcex.cbWndExtra 		= 0;
	wndcex.hInstance 		= hInstance;
	wndcex.hIcon 			= LoadIcon(hInstance, (LPCTSTR)IDI_PROG7_6_1);
	wndcex.hCursor 			= LoadCursor(NULL, IDC_ARROW);
	wndcex.hbrBackground 	= (HBRUSH)(COLOR_WINDOW+1);
	wndcex.lpszMenuName 	= NULL;
	wndcex.lpszClassName 	= "StaticTriangle";
	wndcex.hIconSm 			= LoadIcon(wndcex.hInstance, (LPCTSTR)IDI_PROG7_6_1);

	if(!RegisterClassEx(&wndcex))
		return E_FAIL;

	/* create the window */
	///////////////////////
	g_hInstance = hInstance;

	RECT rectangle = {0, 0, 800, 600};
	AdjustWindowRect(&rectangle, WS_OVERLAPPEDWINDOW, FALSE);

	g_hWindow = CreateWindow(
					"StaticTriangle",
					"Rendering a triangle without any transformations",
					WS_OVERLAPPEDWINDOW,
					CW_USEDEFAULT, CW_USEDEFAULT,
					rectangle.right - rectangle.left,
					rectangle.bottom - rectangle.top,
					NULL, NULL, hInstance, NULL);

	if(!g_hWindow)
		return E_FAIL;

	ShowWindow(g_hWindow, nCmdShow);

	return S_OK;
}

/* create the Direct3D device and swap chain */
HRESULT InitDevice()
{
    HRESULT hresult_ = S_OK;

	/* setup client rectangle width and height */
	/////////////////////////////////////////////
	RECT rectangle;
	GetClientRect(g_hWindow, &rectangle);

	int rectangle_width = rectangle.right - rectangle.left;
	int rectangle_height = rectangle.bottom - rectangle.top;
	int deviceFlags = 0;

	/* specify D3D10 drivers to use */
	/////////////////////////////////
	D3D10_DRIVER_TYPE d3d10drivers[] =
	{
		D3D10_DRIVER_TYPE_HARDWARE,
		D3D10_DRIVER_TYPE_REFERENCE,
	};

	int numberDrivers =	sizeof(d3d10drivers)/sizeof(d3d10drivers[0]);

	/* create the swap chain */
	///////////////////////////
	DXGI_SWAP_CHAIN_DESC swapchain;
	SecureZeroMemory(&swapchain, sizeof(swapchain));

	swapchain.BufferCount = 1;
	swapchain.BufferDesc.Width = rectangle_width;
	swapchain.BufferDesc.Height = rectangle_height;
	swapchain.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
	swapchain.BufferDesc.RefreshRate.Numerator = 30;
	swapchain.BufferDesc.RefreshRate.Denominator = 1;
	swapchain.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
	swapchain.OutputWindow = g_hWindow;
	swapchain.SampleDesc.Count = 1;
	swapchain.SampleDesc.Quality = 0;
	swapchain.Windowed = TRUE;

	for(int d3d10DriverCount = 0; d3d10DriverCount < numberDrivers; d3d10DriverCount++)
	{
		g_dx10DriverType = d3d10drivers[d3d10DriverCount];
		hresult_ = D3D10CreateDeviceAndSwapChain(
						NULL, g_dx10DriverType, NULL,
						deviceFlags, D3D10_SDK_VERSION,
						&swapchain, &g_pdx10SwapChain,
						&g_pd3d10Device);

		if(SUCCEEDED(hresult_))
			break;
	}

	if(FAILED(hresult_))
		return hresult_;

	/* specify the render target view */
	////////////////////////////////////
	ID3D10Texture2D *pBuffer;
	hresult_ = g_pdx10SwapChain->GetBuffer(0, __uuidof(ID3D10Texture2D),(LPVOID*) &pBuffer);

	if(FAILED(hresult_))
		return hresult_;

	hresult_ = g_pd3d10Device->CreateRenderTargetView(pBuffer, NULL, &g_pd3d10RenderTargetView);

	pBuffer->Release();

	if(FAILED(hresult_))
		return hresult_;

	g_pd3d10Device->OMSetRenderTargets(1,&g_pd3d10RenderTargetView, NULL);

	/* create the viewport */
	/////////////////////////
	D3D10_VIEWPORT viewport;
	viewport.TopLeftX 	= 0;
	viewport.TopLeftY 	= 0;
	viewport.Width 		= rectangle_width;
	viewport.Height 	= rectangle_height;
	viewport.MinDepth 	= 0.0f;
	viewport.MaxDepth 	= 1.0f;

	g_pd3d10Device->RSSetViewports(1, &viewport);

	/* load the effect file specifying our triangles color*/
	//////////////////////////////////////////////////////////
	hresult_ = D3DX10CreateEffectFromFile(
					"effect.fx", NULL, NULL, "fx_4_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, 
                    g_pd3d10Device, NULL, NULL, &g_pd3d10Effect, NULL, NULL);

	/* if the effect file is missing */
	///////////////////////////////////
	if(FAILED(hresult_))
	{
		return hresult_;
	}

	/* set the effect technique and input layout */
	///////////////////////////////////////////////
	g_pd3d10EffectTechnique = g_pd3d10Effect->GetTechniqueByName("Triangle");

	D3D10_INPUT_ELEMENT_DESC inputlayout[] =
	{
		{"POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0,
		D3D10_INPUT_PER_VERTEX_DATA, 0},
	};

	int numberOfElements = sizeof(inputlayout) / sizeof(inputlayout[0]);

	D3D10_PASS_DESC PassDescription;
	g_pd3d10EffectTechnique->GetPassByIndex(0)->GetDesc(&PassDescription);

	hresult_ = g_pd3d10Device->CreateInputLayout(
								inputlayout, numberOfElements,
								PassDescription.pIAInputSignature,
								PassDescription.IAInputSignatureSize,
								&g_pd3d10InputVertexLayout);

	if(FAILED(hresult_))
		return hresult_;

	g_pd3d10Device->IASetInputLayout(g_pd3d10InputVertexLayout);

	/* specify our triangles vertices */
	/////////////////////////////////////
	TriangleVertex vertices[] =
	{
		D3DXVECTOR3(0.0f, 1.0f, 1.0f),
		D3DXVECTOR3(1.0f, -1.0f, 1.0f),
		D3DXVECTOR3(-1.0f, -1.0f, 1.0f),
	};

	/* setup the buffer resource */
	///////////////////////////////////
	D3D10_BUFFER_DESC bufferdesc;
	bufferdesc.Usage 			= D3D10_USAGE_DEFAULT;
	bufferdesc.BindFlags 		= D3D10_BIND_VERTEX_BUFFER;
	bufferdesc.CPUAccessFlags 	= 0;
	bufferdesc.MiscFlags 		= 0;
	bufferdesc.ByteWidth 		= sizeof(TriangleVertex)*3;

	/* initialize resource */
	//////////////////////////////
	D3D10_SUBRESOURCE_DATA ResourceData;
	ResourceData.pSysMem = vertices;
	hresult_ = g_pd3d10Device->CreateBuffer(
								&bufferdesc,
								&ResourceData,
								&g_pd3d10VertexBuffer);

	if(FAILED(hresult_))
		return hresult_;

	/* setup the vertex buffer */
	/////////////////////////////
	UINT elementstride = sizeof(TriangleVertex);

	UINT bufferoffset = 0;
	g_pd3d10Device->IASetVertexBuffers(
						0, 1, &g_pd3d10VertexBuffer,
						&elementstride, &bufferoffset);

	/* bind the primitive type and data order information to a
	   triangle list */
	g_pd3d10Device->IASetPrimitiveTopology(
						D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
	return S_OK;
}

/* setup our WndProc callback */
LRESULT CALLBACK WndProc(HWND hWindow, UINT msg, WPARAM wParam, LPARAM lParam)
{
	HDC devicecontext;
	PAINTSTRUCT paintstructure;

	switch(msg)
	{
		case WM_DESTROY:
			PostQuitMessage(0);
			break;

		case WM_PAINT:
			devicecontext = BeginPaint (hWindow, &paintstructure);
			EndPaint(hWindow, &paintstructure);
		break;

	default:
		return DefWindowProc(hWindow, msg, wParam, lParam);
	}
	return 0;
}

/* checks whether an object has been created, and if it has, deallocates it */
void Clean()
{
	if(g_pd3d10Device)
		g_pd3d10Device->ClearState();

	if(g_pd3d10InputVertexLayout)
		g_pd3d10InputVertexLayout->Release();

	if(g_pd3d10VertexBuffer)
		g_pd3d10VertexBuffer->Release();

	if(g_pd3d10RenderTargetView)
		g_pd3d10RenderTargetView->Release();

	if(g_pd3d10Effect)
		g_pd3d10Effect->Release();

	if(g_pd3d10Device)
		g_pd3d10Device->Release();

	if(g_pdx10SwapChain)
		g_pdx10SwapChain->Release();
}

/* draw the triangle */
void DrawTriangle()
{
	/* start by clearing the back-buffer with the color white */
	float ClearBufferColour[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	g_pd3d10Device->ClearRenderTargetView(g_pd3d10RenderTargetView, ClearBufferColour);

	/* obtain the D3D10_TECHNIQUE_DESC effect-variable description */
	D3D10_TECHNIQUE_DESC technique;
	g_pd3d10EffectTechnique->GetDesc(&technique);

	/* render by looping over the number of technique passes */
	for(int i = 0; i < technique.Passes; ++i)
	{
		g_pd3d10EffectTechnique->GetPassByIndex(i)->Apply(0);
		g_pd3d10Device->Draw(3, 0);
	}

	/* switch the back- and front-buffer and display the
	triangle */
	g_pdx10SwapChain->Present(0,0);
}